import markdown

from . import dsref
from . import loader
from . import version_info
from .util import set_fields, build_repr


class Meta(object):
    def __init__(self, obj):
        set_fields(self, obj, ['accessURL', 'accrualPeriodicity', 'citations',
                               'contributors', 'description', 'downloadURL',
                               'homeURL', 'identifier', 'keywords', 'language',
                               'license', 'path' 'readmeURL', 'title', 'theme',
                               'version'])

    def __repr__(self):
        r = build_repr(self)
        return 'Meta(%s)' % r


class Readme(object):
    def __init__(self, obj):
        # TODO(dustmop): Handle scriptPath
        if not obj or 'scriptBytes' not in obj:
            self.script = None
            return
        self.script = loader.base64_decode(obj['scriptBytes']).decode('utf-8')

    def render(self):
        if self.script is None:
            return ''
        return markdown.markdown(self.script)

    def __str__(self):
        if self.script is None:
            return 'None'
        return self.script

    def __repr__(self):
        if self.script is None:
            return 'None'
        return 'Readme("%s")' % self

    def _repr_html_(self):
        return self.render()


class Structure(object):
    def __init__(self, obj):
        set_fields(self, obj, ['checksum', 'depth', 'entries', 'format',
                               'formatConfig', 'length', 'schema'])

    def __repr__(self):
        r = build_repr(self)
        return 'Structure(%s)' % r


class Commit(object):
    def __init__(self, obj):
        set_fields(self, obj, ['author', 'message', 'path', 'signature',
                               'timestamp', 'title'])

    def __repr__(self):
        r = build_repr(self)
        return 'Commit(%s)' % r


class Dataset(object):
    def __init__(self, obj):
        # Fields that are always present
        set_fields(self, obj, ['username', 'name', 'profileID', 'path'])
        if self.username is None and 'peername' in obj:
            self.username = obj.get('peername')
        # Fields returned by `get` commands
        set_fields(self, obj, ['bodyPath', 'previousPath'])
        # Version info
        self.versionInfo = self._build_version_info(obj)
        # Subcomponents
        self.commit_component = Commit(obj.get('commit'))
        self.meta_component = Meta(obj.get('meta'))
        self.readme_component = Readme(obj.get('readme'))
        self.structure_component = Structure(obj.get('structure'))
        self.body_component = None

    def _build_version_info(self, obj):
        info = version_info.VersionInfo(obj)
        # TODO(dustmop): Remove me after this typo is fixed in qri core
        if 'bodyFromat' in obj:
            info.body_format = obj.get('bodyFromat')
        return info

    @property
    def meta(self):
        return self.meta_component

    @property
    def commit(self):
        return self.commit_component

    @property
    def readme(self):
        return self.readme_component

    @property
    def structure(self):
        return self.structure_component

    @property
    def body(self):
        if self.structure_component is None:
            raise RuntimeError('Cannot read body without structure')
        if self.structure.format != 'csv':
            raise RuntimeError('Only csv body format is supported')
        if self.body_component is None:
            ref = dsref.Ref(self.username, self.name)
            df = loader.instance().load_body(ref, self.structure)
            self.body_component = df
        return self.body_component

    def human_ref(self):
        return '%s/%s' % (self.username, self.name)

    def __repr__(self):
        return 'Dataset("%s")' % self.human_ref()

    def _repr_html_(self):
        return '<code>Dataset("%s")</code>' % self.human_ref()


class DatasetList(list):
    def __repr__(self):
        content = ', '.join(['%s' % d for d in self])
        return '[%s]' % (content,)

    def _repr_html_(self):
        content = ', '.join(['%s' % d for d in self])
        return '<code>[%s]</code>' % (content,)



import base64
import io
import json
import pandas

from .cmd_util import shell_exec, QriClientError
import requests
from subprocess import Popen, PIPE


_inst = None


def instance():
    global _inst
    if _inst is None:
        print
        proc = Popen(['which', 'qri'], stdout=PIPE)
        stdout, err = proc.communicate()
        if proc.returncode == 0:
            # Have a local qri binary
            _inst = LocalQriBinaryRepo()
        else:
            # Send http requests to api.qri.cloud
            _inst = CloudAPIRepo()
    return _inst


def set_instance(obj):
    global _inst
    _inst = obj


class LocalQriBinaryRepo(object):
    def get_dataset_object(self, ref):
        cmd = 'qri get --format json %s' % ref.human()
        result, err = shell_exec(cmd)
        if err:
            raise QriClientError(err)
        return json.loads(result)

    def list_dataset_objects(self, username=None):
        if username is not None:
            raise QriClientError('listing with username not supported')
        cmd = 'qri list --format json'
        result, err = shell_exec(cmd)
        if err:
            raise QriClientError(err)
        return json.loads(result)

    def pull_dataset(self, ref):
        cmd = 'qri pull %s' % ref.human()
        result, err = shell_exec(cmd)
        if err:
            raise QriClientError(err)
        return result

    def load_body(self, ref, structure):
        if structure.format != 'csv':
            raise RuntimeError('Format "%s" not supported' % structure.format)
        cmd = 'qri get body %s' % ref.human()
        result, err = shell_exec(cmd)
        if err:
            raise QriClientError(err)
        stream = io.StringIO(str(result, 'utf8'))
        columns = [e for e in structure.schema['items']['items']]
        col_names = [c['title'] for c in columns]
        types = {c['title']: pd_type(c['type']) for c in columns}
        header = 0 if structure.format_config.get('headerRow') else None
        df = None
        try:
            # Try to parse the csv using the schema
            df = pandas.read_csv(stream, header=header, names=col_names,
                                 dtype=types)
        except (TypeError, ValueError):
            # If pandas encountered parse errors, reparse without datatypes
            stream = io.StringIO(str(result, 'utf8'))
            df = pandas.read_csv(stream, header=header, names=col_names)
        return df


class CloudAPIRepo(object):
    def get_dataset_object(self, ref):
        r = requests.get('https://api.qri.cloud/get/%s' % ref.human())
        return json.loads(r.text)['data']['dataset']

    def list_dataset_objects(self, username=None):
        raise NotImplementedError('CloudAPIRepo.list_dataset_objects')

    def load_body(self, ref, structure):
        raise NotImplementedError('CloudAPIRepo.load_body_dataframe')


def from_json(json_text):
    return pandas.read_json(json_text)


def base64_decode(bdata):
    return base64.b64decode(bdata)


def pd_type(t):
    if t == 'integer':
        return 'int64'
    elif t == 'number':
        return 'float64'
    elif t == 'string':
        return 'string'
    elif t == 'bool':
        return 'bool'
    raise RuntimeError('Unknown type: "%s"' % t)



from .util import set_fields


class VersionInfo(object):
    def __init__(self, obj):
        set_fields(self, obj, ['bodySize', 'bodyRows', 'bodyFormat', 'numErrors', 'commitTime'])

    def __repr__(self):
        return 'VersionInfo()'



import re

from .cmd_util import QriClientError


class Ref(object):
    def __init__(self, username, name):
        self.username = username
        self.name = name

    def human(self):
        return '{}/{}'.format(self.username, self.name)

    def clone(self):
        return Ref(self.username, self.name)

    def __str__(self):
        return self.human()


def parse_ref(s):
    m = re.match(r'([a-z][a-z0-9_-]*)/([a-z][a-z0-9_-]*)', s)
    if m:
        username = m.group(1)
        name = m.group(2)
        return Ref(username, name)
    raise QriClientError('Could not parse reference "{}"'.format(s))



import re


def to_snake_case(text):
    return re.sub(r'([^A-Z])([A-Z])', r'\1_\2', text).lower()


def set_fields(inst, obj, fields):
    field_list = []
    for f in fields:
        name = to_snake_case(f)
        field_list.append(name)
        if obj is None:
            setattr(inst, name, None)
        else:
            setattr(inst, name, obj.get(f))
    setattr(inst, '_fields', field_list)


def build_repr(inst):
    accum = []
    for f in inst._fields:
        val = getattr(inst, f)
        if val:
            accum.append('%r: %r' % (f, val))
    return '{' + ', '.join(accum) + '}'



import re
import shlex
from subprocess import Popen, PIPE
import sys


def shell_exec(command, cwd=None):
    """execute commands and return stdout"""
    if isinstance(command, list):
        command_list = command
    else:
        command_list = shlex.split(command)
    try:
        proc = Popen(command_list,
                     stdin=PIPE,
                     stdout=PIPE,
                     stderr=PIPE,
                     cwd=cwd)
        stdout, err = proc.communicate()
        if proc.returncode == 0:
            # If command exit code is 0, assume stderr is informational only.
            # This probably won't work forever, but fits most of our current
            # use cases.
            err = None
        return stdout, err
    except FileNotFoundError:
        sys.stderr.write("""qri command-line binary not found. It is either not installed, or PATH needs to be assigned. Please get the latest release from https://github.com/qri-io/qri, then run this command again.\n""")
        sys.exit(1)
    except BaseException as e:
        raise e


def strip_color(colored_text):
    """strips color characters to return plaintext for when the
       global color flag cannot be overriden"""
    ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', colored_text)


class QriClientError(RuntimeError):
    def __init__(self, err_string):
        clean_err_string = strip_color(err_string.decode('utf-8'))
        super(QriClientError, self).__init__(clean_err_string)



#!/usr/bin/python

"""Client for interacting with qri repositories"""

from .cmd_util import shell_exec, QriClientError
from . import dataset
from . import dsref
from . import loader


def list(username=None):
    """list datasets in the user's repository"""
    objs = loader.instance().list_dataset_objects(username)
    datasets = dataset.DatasetList([dataset.Dataset(d) for d in objs])
    datasets.sort(key=lambda d: d.human_ref())
    return datasets


def get(refstr):
    """get a dataset in the repository by reference"""
    ref = dsref.parse_ref(refstr)
    obj = loader.instance().get_dataset_object(ref)
    d = dataset.Dataset(obj)
    return d


def pull(refstr):
    """pull a remote dataset from the registry to the user's repository"""
    ref = dsref.parse_ref(refstr)
    print('Fetching from registry...')
    text = loader.instance().pull_dataset(ref)
    return 'Pulled %s: %s' % (ref, text)


def add(refstr):
    """add is an alias for pull"""
    return pull(refstr)


def sql(query):
    """sql query run against a dataset"""
    cmd = ['qri', 'sql', '--format', 'json', query]
    result, err = shell_exec(cmd)
    if err:
        raise QriClientError(err)
    return loader.from_json(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools

with open('README.md', 'r') as fp:
    long_description = fp.read()
    pos = long_description.find('# Development')
    if pos > -1:
        long_description = long_description[:pos]

setuptools.setup(
    name='qri',
    version='0.1.4',
    author='Dustin Long',
    author_email='dustmop@qri.io',
    description='qri python client',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/qri-io/qri-python',
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas==1.0.0',
        'Markdown==3.2.2',
        'requests==2.24.0',
    ],
    classifiers=[
        'Programming Language :: Python',
        'License :: OSI Approved :: MIT License',
    ],
    python_requires='>=3.6'
)
